package core

import (
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/cli/q/src/expression"
	"git.akyoto.dev/cli/q/src/types"
)

// CompileCall executes a function call.
// All call registers must hold the correct parameter values before the function invocation.
// Registers that are in use must be saved if they are modified by the function.
// After the function call, they must be restored in reverse order.
func (f *Function) CompileCall(root *expression.Expression) ([]types.Type, error) {
	var (
		pkg      = f.Package
		pkgNode  *expression.Expression
		name     string
		nameNode = root.Children[0]
		fn       *Function
		exists   bool
	)

	if nameNode.IsLeaf() {
		name = nameNode.Token.Text(f.File.Bytes)

		switch name {
		case "len":
			return _len.OutputTypes, f.CompileLen(root)
		case "syscall":
			return nil, f.CompileSyscall(root)
		case "new":
			typ, err := f.CompileNew(root)
			return []types.Type{typ}, err
		case "delete":
			return nil, f.CompileDelete(root)
		case "store":
			return nil, f.CompileMemoryStore(root)
		}
	} else {
		pkgNode = nameNode.Children[0]
		nameNode = nameNode.Children[1]

		pkg = pkgNode.Token.Text(f.File.Bytes)
		name = nameNode.Token.Text(f.File.Bytes)
	}

	if f.UniqueName == "core.init" && pkg == "main" && name == "main" {
		f.Call("main.main")
		return nil, nil
	}

	fn, exists = f.All.Functions[pkg+"."+name]

	if !exists {
		return nil, errors.New(&errors.UnknownFunction{Name: name}, f.File, nameNode.Token.Position)
	}

	if pkg != f.File.Package && !fn.IsExtern() {
		if f.File.Imports == nil {
			return nil, errors.New(&errors.UnknownPackage{Name: pkg}, f.File, pkgNode.Token.Position)
		}

		imp, exists := f.File.Imports[pkg]

		if !exists {
			return nil, errors.New(&errors.UnknownPackage{Name: pkg}, f.File, pkgNode.Token.Position)
		}

		imp.Used = true
	}

	parameters := root.Children[1:]

	if len(parameters) != len(fn.Input) {
		return nil, errors.New(&errors.ParameterCountMismatch{Function: fn.Name, Count: len(parameters), ExpectedCount: len(fn.Input)}, f.File, nameNode.Token.End())
	}

	if fn.IsExtern() {
		return f.CallExtern(fn, parameters)
	}

	registers := f.CPU.Input[:len(parameters)]
	err := f.ExpressionsToRegisters(parameters, registers, fn.Input, true)

	if err != nil {
		return nil, err
	}

	f.CallSafe(fn, registers)
	return fn.OutputTypes, nil
}
